const path = require('path');
const mkdirp = require('mkdirp');
const urlJoin = require('url-join');
const Generator = require('yeoman-generator');
const camelize = require('underscore.string/camelize');
const capitalize = require('underscore.string/capitalize');
const underscored = require('underscore.string/underscored');
const recast = require('recast');
const _ = require('lodash');
const rimraf = require('rimraf');

module.exports = class extends Generator {
  constructor(args, opts) {
    super(args, opts);

    this.projectname = this.config.get('name');
    this.apiversion = this.config.get('apiversion');
  }

  prompting() {
    const done = this.async();
    const that = this;

    this.prompt([
      {
        type: 'input',
        name: 'name',
        message: 'What is your GraphQL query name?',
        default: this.config.get('last_endpoint') || 'endpoint',
        validate: value => value !== undefined && value !== ''
      },
      {
        type: 'input',
        name: 'apidesc',
        message: 'Please give API description for documentation!',
        default: this.config.get('last_apidesc') || 'description'
      }
    ]).then((answers) => {
      that.props = answers;
      that.props.injectRoute = false;
      that.props.apibase = that.config.get('apibase') || 'api';
      that.props.apiversion = that.config.get('apiversion') || 'v1';
      that.props.endpoint = `${that.props.name}`;
      that.props.filename = underscored(that.props.name).replace('_', '-');
      that.props.src = that.config.get('src') || 'src';
      that.props.apiPath = urlJoin(that.props.src, that.props.apibase);
      that.props.kind = that.config.get('kind');
      that.props.projectname = that.projectname;
      that.props.author = that.config.get('author');
      that.props.date = new Date().toLocaleString("en-US");

      // Handle nested api path
      let { name } = that.props;

      // Dirname by default is the filename
      let dirname = that.props.filename;

      if (name.indexOf('/') > -1) {
        // Create dirname
        dirname = underscored(path.dirname(name)).replace('_', '-');
        // mkdirp.sync(path.join(this.destinationPath(), dirname));
        name = path.basename(name);

        // now filename only store the basename
        that.props.filename = underscored(name).replace('_', '-');

        // dirname should be full path
        dirname = urlJoin(dirname, that.props.filename);
      }
      that.props.dirname = dirname;
      that.props.name = camelize(name, true);
      that.props.gqlResolveName = dirname.split('/').map(name => capitalize(name)).join('').replace('-', '');

      if (!that.projectname || !that.apiversion || that.props.kind !== 'hybrid') {
        console.log('Invalid project type! exiting');
        process.exit();
      }

      this.on('end', () => {
        this.config.set('last_endpoint', that.props.endpoint);
        this.config.set('last_apidesc', that.props.apidesc);
        this.config.set('apibase', that.props.apibase);
        this.config.set('apiversion', that.props.apiversion);
      });

      done();
    });
  }

  writing() {
    const done = this.async();
    const { props } = this;
    props.src = this.config.get('src') || 'src';
    const copyTpl = this.fs.copyTpl.bind(this.fs);
    const tPath = this.templatePath.bind(this);
    const dPath = this.destinationPath.bind(this);
    props._name_ = props.name.toLowerCase();
    const apiRootPath = urlJoin(props.apiPath, this.apiversion);
    const apiPath = `${apiRootPath}/${props.dirname}`;

    // Template
    const deliveryPath = 'delivery';
    const schemaPath = 'schema';
    const useCasesPath = 'usecases';

    // Sync the components path directory
    mkdirp.sync(path.join(this.destinationPath(), apiPath));

    const targetPaths = {
      index: {
        source: '_index.ejs',
        target: urlJoin(apiPath, 'index.js')
      },
      spec: {
        source: urlJoin(useCasesPath, '_spec.ejs'),
        target: urlJoin(apiPath, useCasesPath, `${props.filename}.spec.js`)
      },
      usecase: {
        source: urlJoin(useCasesPath, '_usecase.ejs'),
        target: urlJoin(apiPath, useCasesPath, `${props.filename}.usecase.js`)
      }
    };

    const copyFiles = (destPaths) => {
      Object.keys(destPaths).forEach((fileKey) => {
        const file = destPaths[fileKey];
        const sourcePath = tPath(file.source);
        const destinationPath = dPath(file.target);
        if (!this.fs.exists(destinationPath)) {
          copyTpl(sourcePath, destinationPath, props);
        }
      });  
    }

    copyFiles(targetPaths);

    const gqlPaths = {
      resolver: {
        source: urlJoin(deliveryPath, '_gql.resolver.ejs'),
        target: urlJoin(apiPath, deliveryPath, `${props.filename}.gql.resolver.js`)
      },
      schema: {
        source: urlJoin(schemaPath, '_gql.schema.ejs'),
        target: urlJoin(apiPath, schemaPath, `${props.filename}.gql.schema.js`)
      },
    }

    if (this.props.kind === 'hybrid') {
      copyFiles(gqlPaths);
    }

    done();
  }

  injectGQLSchemaAndResolver() {
    if (this.props.kind === 'hybrid') {
      this._injectGraphQL(urlJoin('src', 'config', 'graphql', 'schema.js'), 'schema', 'Schema');
      this._injectGraphQL(urlJoin('src', 'config', 'graphql', 'resolvers.js'), 'resolver', 'Resolver');
    }
  }

  _injectGraphQL(targetFile, prefix, suffix) {
    this.props.src = this.config.get('src') || 'src';
    const { props } = this;
    const dPath = this.destinationPath.bind(this);

    const indexFile = dPath(targetFile);

    /**
     * Inject codes
     */

    if (this.fs.exists(indexFile)) {
      let ast = recast.parse(this.fs.read(indexFile));
      let { body } = ast.program;

      let injectRequire = true;

      /**
       * Route require injection to index.js
       */
      recast.visit(ast, {
        visitVariableDeclaration(path) {  // eslint-disable-line
          const { declarations } = path.node;

          // Check for const = require variable declaration
          if (declarations.length > 0) {
            const decl = declarations[0];
            const arg = decl.init.arguments.length > 0 ? decl.init.arguments[0].value : '';
            if (arg && arg.trim() === `./${props.name}`) {
              injectRequire = false;
              return injectRequire;
            }
          }
          this.traverse(path);
        }
      });

      // Inject const = require()
      let lastImportIndex;
      if (injectRequire) {
        lastImportIndex = _.findLastIndex(body, statement => statement.type === 'VariableDeclaration');
        const actualImportCode = recast.print(body[lastImportIndex]).code;
        const importString = [`const { ${prefix}: `, props.gqlResolveName, suffix, ' } = require(\'', urlJoin(`@${props.apibase}`, props.apiversion, props.dirname), '\');'].join('');

        body.splice(lastImportIndex < 0 ? 0 : lastImportIndex, 1, importString);
        body.splice(lastImportIndex < 0 ? 0 : lastImportIndex, 0, actualImportCode);
      }

      /**
       * Inject router.use() expression
       */
      let elements = [];
      const lastExportIndex = _.findLastIndex(body, (statement) => { // eslint-disable-line
        if (statement.type === 'ExpressionStatement' && statement.expression && statement.expression.right.type === 'ArrayExpression') {
          elements = statement.expression.right.elements; // eslint-disable-line
          return true;
        }
      });

      // The value we want to inject
      const injectedValue = `${props.gqlResolveName}${suffix}`;

      // We first create the declaration
      let exportArray = ['module.exports = [\n'];

      // Readd current boot array
      elements.forEach((el) => {
        if (el.name !== `${injectedValue}`) {
          exportArray.push(`\t${el.name},\n`);
        }
      });

      // Close our array
      exportArray.push(`\t${injectedValue}\n`, '];');
      const exportString = exportArray.join('');
      
      if (lastExportIndex === -1) {
        const exportRouterIndex = _.findIndex(body, statement => statement.type === 'ExpressionStatement' && statement.expression.right.type === 'ArrayExpression');
        body.splice(exportRouterIndex < 0 ? lastImportIndex : exportRouterIndex, 0, exportString);
      } else {
        // Delete the current array
        body.splice(lastExportIndex, 1);

        // Add the new array
        body.splice(lastExportIndex < 0 ? (lastImportIndex + 1) : lastExportIndex, 0, exportString);
      }

      if (injectRequire) {
        rimraf(indexFile, () => {
          this.fs.write(indexFile, recast.print(ast).code);
        });
      }
    }
  }
};
